/*
 *  $Id: chunkHandler.c,v 1.7 2001/03/05 10:03:32 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 *
 * Purpose: This module assembles and disassembles control chunks.
 *          When a chunk is created, the caller gets a chunks-ID, with which it can
 *          address the chunk in following calls.
 *          In order to handle more than one chunk at the same time, pointers to chunk
 *          are stored in array of length MAX_CHUNKS. This is required for modules that
 *          receive a chunk and want to respond by sending another chunk.
 *          But a chunk should be deleted after a signal is handled. Else the number of
 *          could not be limited and a list would be required to store them.
 *          As long as the bytestring is not read by ch_chunkString, the length of a chunk is
 *          is stored in host byte order, because it may still change as new elements are
 *          added to the chunk.
 *          After ch_chunkString is called, the chunk should no longer be modified.
 *
 *    TODO :   This module also contains a whole lot of malloc()s, and these are always
 *              freed after the chunks has been dealt with. Why not alloc once, and keep
 *              the memory afterwards ????????
 * function prefix: ch_
 */

#include <errno.h>
#include <stdlib.h>             /* for malloc() */

#include "sctp.h"
#include "globals.h"
#include "auxiliary.h"
#include "chunkHandler.h"
#include "SCTP-control.h"
#include "md5.h"


#define MAX_CHUNKS 8

/* Other constants */


static unsigned short writeCursor[MAX_CHUNKS];
static SCTP_simple_chunk *chunks[MAX_CHUNKS];
static boolean chunkCompleted[MAX_CHUNKS];

static ChunkID freeChunkID = 0;


/******************************* internal functions ***********************************************/


/*
 * computes a cookie signature.
 * TODO: replace this by something safer than MD5
 */
static int
signCookie(unsigned char *cookieString, unsigned short cookieLength,
           unsigned char *start_of_signature)
{
    int i;
    MD5_CTX ctx;
    SCTP_our_cookie *cookie;
    unsigned char * key;

    if (cookieString == NULL)
        return -1;
    if (start_of_signature == NULL)
        return -1;
    if (cookieLength == 0)
        return -1;
    key =  key_operation(KEY_READ);
    if (key == NULL)
        return -1;

    cookie = (SCTP_our_cookie *) cookieString;
    memset(cookie->hmac, 0, HMAC_LEN);

    MD5Init(&ctx);
    MD5Update(&ctx, cookieString, cookieLength);
    MD5Update(&ctx, key, SECRET_KEYSIZE);
    MD5Final(start_of_signature, &ctx);

    event_log(INTERNAL_EVENT_0, "Computed MD5 signature : ");
    for (i = 0; i < 4; i++) {
        event_logiiii(VERBOSE, "%2.2x %2.2x %2.2x %2.2x",
                      start_of_signature[i * 4], start_of_signature[i * 4 + 1],
                      start_of_signature[i * 4 + 2], start_of_signature[i * 4 + 3]);
    }

    return 0;
}


/**
 * retrieveVLParamFromString scans for a parameter of a certain type in a message string.
 * The message string must point to a parameter header. The function can also be used
 * to find parameters within a parameter (e.g. addresses within a cookie).
 * @param paramType type of paramter to scan for,
 * @param mstring   pointer to the first parameter header, from which we start scanning
 * @param length    maximum length of parameter field, that may be scanned.
 * @return          position of first parameter occurence, relative to where mstring pointed to
 *                  i.e. 0 returned, when mstring points to the parameter we scan for.
 *                  OR -1 if not found !!!!!!!
 */
static gint32 retrieveVLParamFromString(guint16 paramType, guchar * mstring, guint16 length)
{
    guint16 curs;
    guint16 pType;
    SCTP_vlparam_header *param_header;

    curs = 0;

    /* TODO : add support for all Error Cause Codes ? ECC_XXXXXXXX */
    while (curs < length) {
        param_header = (SCTP_vlparam_header *) & mstring[curs];
        pType = ntohs(param_header->param_type);

        if (pType == paramType) {
            return curs;
        }
            else if (pType == VLPARAM_IPV4_ADDRESS ||
                     pType == VLPARAM_IPV6_ADDRESS ||
                     pType == VLPARAM_COOKIE ||
                     pType == VLPARAM_COOKIE_PRESERV ||
                     pType == ECC_STALE_COOKIE_ERROR || pType == VLPARAM_SUPPORTED_ADDR_TYPES) {
            curs += ntohs(param_header->param_length);
            /* take care of padding */
            while ((curs % 4) != 0)
                curs++;
        } else if (pType == VLPARAM_ECN_CAPABLE || pType == VLPARAM_HOST_NAME_ADDR) {
            event_logi(INTERNAL_EVENT_1, "parameter type %d not supported", pType);
            curs += ntohs(param_header->param_length);
            /* take care of padding here */
            while ((curs % 4) != 0)
                curs++;
        } else {
            error_logi(ERROR_MINOR, "unknown parameter type %u in message", pType);
            if (ntohs(param_header->param_length) == 0) {
                error_log(ERROR_MINOR, "Program/Peer implementation problem : parameter length 0");
                return -1;
            }
            /* try to continue parsing */
            if ((ntohs(param_header->param_length) + curs) < length)
                curs += ntohs(param_header->param_length);
            else
                break;
            /* take care of padding here */
            while ((curs % 4) != 0)
                curs++;
        }
    }
    return -1;
}


/**
 * retrieveIPv4AddressesFromString finds all IPv4 addresses in a message string.
 * TODO : check for maximum number of addresses, there may possibly be an overflow
 * @param   mstring     pointer to the beginning of the message string
 * @param   length      maximum length of the message that is scanned for addresses
 * @param   addresses   field in which we may return addresses
 * @return  number of IPv4 addresses found
*/
static gint32
retrieveIPAddressesFromString(unsigned char *mstring, guint16 length, union sockunion addresses[])
{
    gint32 cursabs;
    gint32 cursrel;
    SCTP_ip_address *address;
    int nAddresses;

    cursabs = 0;
    cursrel = 0;
    nAddresses = 0;

    event_logi(VERBOSE, "retrieveIPAddressesFromString : length = %u", length);

    while (
           (cursrel =
            retrieveVLParamFromString(VLPARAM_IPV4_ADDRESS, &mstring[cursabs],
                                      length - cursabs)) >= 0) {
        address = (SCTP_ip_address *) & mstring[cursabs + cursrel];

        if (IS_IPV4_ADDRESS_PTR_NBO(address)) {
            /* FIXME : either NBO or HBO -- do not mix these */
            addresses[nAddresses].sa.sa_family = AF_INET;
            addresses[nAddresses].sin.sin_port = 0;
            addresses[nAddresses].sin.sin_addr.s_addr = address->dest_addr.sctp_ipv4;
            nAddresses++;
            event_logi(VERBOSE, "Found IPv4 Address = %x", address->dest_addr.sctp_ipv4);

        } else {

            error_log(ERROR_MAJOR,
                      "parameter problem, abort scanning in retrieveIPAddressesFromString");
            break;

        }
        cursabs += cursrel;
        cursabs += 8;
        if (cursabs >= length)
            return nAddresses;
    }

#ifdef HAVE_IPV6
    /* and scan again from the very beginning................. */
    cursabs = 0;
    cursrel = 0;
    while (
           (cursrel =
            retrieveVLParamFromString(VLPARAM_IPV6_ADDRESS, &mstring[cursabs],
                                      length - cursabs)) >= 0) {
        address = (SCTP_ip_address *) & mstring[cursabs + cursrel];

        if (IS_IPV6_ADDRESS_PTR_NBO(address)) {
            addresses[nAddresses].sa.sa_family = AF_INET6;
            addresses[nAddresses].sin6.sin6_port = 0;
            memcpy(addresses[nAddresses].sin6.sin6_addr.s6_addr,
                   address->dest_addr.sctp_ipv6, sizeof(struct in6_addr));
            nAddresses++;
            event_log(VERBOSE, "Found IPv6 Address !");

        } else {

            error_log(ERROR_MAJOR,
                      "parameter problem, abort scanning in retrieveIPv4AddressesFromString");
            break;

        }
        cursabs += cursrel;
        cursabs += 20;
        if (cursabs >= length)
            break;
    }
#endif

    return nAddresses;
}


static void enterChunk(SCTP_simple_chunk * chunk, char *log_text)
{
    unsigned int cid;

    freeChunkID = (freeChunkID + 1) % MAX_CHUNKS;

    cid = freeChunkID;
    event_logi(INTERNAL_EVENT_0, log_text, cid);

    chunks[freeChunkID] = chunk;
    writeCursor[freeChunkID] = 0;
    chunkCompleted[freeChunkID] = FALSE;
}

/******************************* external functions ***********************************************/

/******create, write into and read from init and initAck ******************************************/

/* ch_makeInit makes an init and initializes the the fixed part of init */
ChunkID
ch_makeInit(unsigned int initTag,
            unsigned int rwnd,
            unsigned short noOutStreams, unsigned short noInStreams, unsigned int initialTSN)
{
    SCTP_init *initChunk;

    /* creat init chunk */
    initChunk = (SCTP_init *) malloc(sizeof(SCTP_init));
    if (initChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    /* enter fixed part of init */
    initChunk->chunk_header.chunk_id = CHUNK_INIT;
    initChunk->chunk_header.chunk_flags = 0x00;
    initChunk->chunk_header.chunk_length = sizeof(SCTP_chunk_header) + sizeof(SCTP_init_fixed);
    initChunk->init_fixed.init_tag = htonl(initTag);
    initChunk->init_fixed.rwnd = htonl(rwnd);
    initChunk->init_fixed.outbound_streams = htons(noOutStreams);
    initChunk->init_fixed.inbound_streams = htons(noInStreams);
    initChunk->init_fixed.initial_tsn = htonl(initialTSN);

    enterChunk((SCTP_simple_chunk *) initChunk, "created init %u ");

    return freeChunkID;
}



/* ch_makeInitAck makes an initAck and initializes the the fixed part of initAck */
ChunkID
ch_makeInitAck(unsigned int initTag,
               unsigned int rwnd,
               unsigned short noOutStreams, unsigned short noInStreams, unsigned int initialTSN)
{
    SCTP_init *initAckChunk;

    /* creat init chunk */
    initAckChunk = (SCTP_init *) malloc(sizeof(SCTP_init));
    if (initAckChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    /* enter fixed part of init */
    initAckChunk->chunk_header.chunk_id = CHUNK_INIT_ACK;
    initAckChunk->chunk_header.chunk_flags = 0x00;
    initAckChunk->chunk_header.chunk_length = sizeof(SCTP_chunk_header) + sizeof(SCTP_init_fixed);
    initAckChunk->init_fixed.init_tag = htonl(initTag);
    initAckChunk->init_fixed.rwnd = htonl(rwnd);
    initAckChunk->init_fixed.outbound_streams = htons(noOutStreams);
    initAckChunk->init_fixed.inbound_streams = htons(noInStreams);
    initAckChunk->init_fixed.initial_tsn = htonl(initialTSN);

    enterChunk((SCTP_simple_chunk *) initAckChunk, "created initAckChunk %u ");

    return freeChunkID;
}

void
ch_enterSupportedAddressTypes(ChunkID chunkID,
                              gboolean with_ipv4, gboolean with_ipv6, gboolean with_dns)
{
    SCTP_supported_addresstypes *param = NULL;
    guint16 num_of_types = 0, position = 0;
    guint16 total_length = 0;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
    }

    if (chunkCompleted[chunkID]) {
        error_log(ERROR_MAJOR, " ch_enterSupportedAddressTypes : chunk already completed");
        return;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        if (with_ipv4)
            num_of_types = 1;
        if (with_ipv6)
            num_of_types++;
        if (with_dns)
            num_of_types++;

        /* append the new parameter */
        param = (SCTP_supported_addresstypes *) & ((SCTP_init *)
                                                   chunks[chunkID])->variableParams
            [writeCursor[chunkID]];
        /* _might_ be overflow here, at some time... */
        if (num_of_types == 0)
            error_log(ERROR_FATAL, " No Supported Address Types -- Program Error");

        total_length = sizeof(SCTP_vlparam_header) + num_of_types * sizeof(guint16);

        writeCursor[chunkID] += total_length;
        if ((total_length % 4) != 0)
            writeCursor[chunkID] += 2;

        /* enter cookie preservative */
        param->vlparam_header.param_type = htons(VLPARAM_SUPPORTED_ADDR_TYPES);
        param->vlparam_header.param_length = htons(total_length);
        if (with_ipv4) {
            param->address_type[position] = htons(VLPARAM_IPV4_ADDRESS);
            position++;
        }
        if (with_ipv6) {
            param->address_type[position] = htons(VLPARAM_IPV6_ADDRESS);
            position++;
        }
        if (with_dns) {
            param->address_type[position] = htons(VLPARAM_HOST_NAME_ADDR);
            position++;
        }
        /* take care of padding */
        if (position == 1 || position == 3)
            param->address_type[position] = htons(0);

    } else {
        error_log(ERROR_MAJOR, "ch_enterSupportedAddressTypes : chunk type not init");
    }
}


/**
 * ch_enterCookiePreservative appends a cookie preservative with the suggested
 * cookie lifespan to an init chunk.
 */
void ch_enterCookiePreservative(ChunkID chunkID, unsigned int lifespanIncrement)
{
    SCTP_cookie_preservative *preserv;
    gint32 vl_param_curs;
    guint16 vl_param_total_length;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
    }

    if (chunkCompleted[chunkID]) {
        error_log(ERROR_MAJOR, "ch_enterCookiePreservative : chunk already completed");
        return;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        /* check if init chunk already contains a cookie preserv. */
        vl_param_total_length =
            ((SCTP_init *) chunks[chunkID])->chunk_header.chunk_length -
            sizeof(SCTP_chunk_header) - sizeof(SCTP_init_fixed);

        vl_param_curs = retrieveVLParamFromString(VLPARAM_COOKIE_PRESERV, &((SCTP_init *)
                                                                            chunks
                                                                            [chunkID])->
                                                  variableParams[0], vl_param_total_length);
        if (vl_param_curs >= 0) {
            /* simply overwrite this cookie preserv. */
            preserv = (SCTP_cookie_preservative *) & ((SCTP_init *)
                                                      chunks[chunkID])->variableParams
                [vl_param_curs];
        } else {
            /* append the new parameter */
            preserv = (SCTP_cookie_preservative *) & ((SCTP_init *)
                                                      chunks[chunkID])->variableParams
                [writeCursor[chunkID]];
            /* _might_ be overflow here, at some time... */
            writeCursor[chunkID] += sizeof(SCTP_cookie_preservative);
        }

        /* enter cookie preservative */
        preserv->vlparam_header.param_type = htons(VLPARAM_COOKIE_PRESERV);
        preserv->vlparam_header.param_length = htons(sizeof(SCTP_cookie_preservative));
        preserv->cookieLifetimeInc = htonl(lifespanIncrement);

    } else {
        error_log(ERROR_MAJOR, "ch_enterCookiePreservative: chunk type not init");
    }
}



/**
 *  ch_enterIPaddresses appends local IP addresses to a chunk designated by
 *  chunkID, usually an init or initAck
 *
 *  TODO : Check this........!!
 */
void ch_enterIPaddresses(ChunkID chunkID, union sockunion sock_addresses[], int noOfAddresses)
{
    unsigned char *mstring;
    int i;
    int length;
    SCTP_ip_address *address;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        if (chunkCompleted[chunkID]) {
            error_log(ERROR_MAJOR, "ch_enterIPaddresses: chunk already completed");
            return;
        }

        length = 0;
        mstring = &((SCTP_init *) chunks[chunkID])->variableParams[writeCursor[chunkID]];

        for (i = 0; i < noOfAddresses; i++) {
            address = (SCTP_ip_address *) & mstring[length];

            switch (sockunion_family(&(sock_addresses[i]))) {
            case AF_INET:
                address->vlparam_header.param_type = htons(VLPARAM_IPV4_ADDRESS);
                address->vlparam_header.param_length = htons(8);
                address->dest_addr.sctp_ipv4 = sock2ip(&(sock_addresses[i]));
                length += 8;
                break;
#ifdef HAVE_IPV6
            case AF_INET6:
                address->vlparam_header.param_type = htons(VLPARAM_IPV6_ADDRESS);
                address->vlparam_header.param_length = htons(20);
                memcpy(address->dest_addr.sctp_ipv6,
                       &(sock2ip6(&(sock_addresses[i]))), sizeof(struct in6_addr));
                length += 20;
                break;
#endif
            default:
                error_logi(ERROR_MAJOR, "Unsupported Address Family %d",
                           sockunion_family(&(sock_addresses[i])));
                break;

            }
        }
        writeCursor[chunkID] += length;
    } else {
        error_log(ERROR_MAJOR, "ch_enterIPaddresses: chunk type not init or initAck");
    }

    return;
}



/* ch_enterCookieVLP enters the variable length params of cookie into an initAck */
void
ch_enterCookieVLP(ChunkID chunkID,
                  SCTP_init_fixed * init_fixed,
                  SCTP_init_fixed * initAck_fixed,
                  guint32 cookieLifetime,
                  guint32 local_tie_tag,
                  guint32 peer_tie_tag,
                  union sockunion local_Addresses[],
                  guint16 num_local_Addresses,
                  union sockunion peer_Addresses[], guint16 num_peer_Addresses)
{
    SCTP_cookie_param *cookie;
    unsigned short wCurs;
    int result, count;
    guint16 no_local_ipv4_addresses = 0;
    guint16 no_remote_ipv4_addresses = 0;
    guint16 no_local_ipv6_addresses = 0;
    guint16 no_remote_ipv6_addresses = 0;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK) {
        if (chunkCompleted[chunkID]) {
            error_log(ERROR_MAJOR, "ch_enterCookieVLP: chunk already completed");
            return;
        }

        /* enter fixed length params into cookie (which is variable part of initAck) */
        cookie = (SCTP_cookie_param *)
            & ((SCTP_init *) chunks[chunkID])->variableParams[writeCursor[chunkID]];
        cookie->vlparam_header.param_type = htons(VLPARAM_COOKIE);
        /* these contain the CURRENT tags ! */
        cookie->z_side_initAck = *initAck_fixed;
        cookie->a_side_init = *init_fixed;

        cookie->local_tie_tag = htonl(local_tie_tag);
        cookie->peer_tie_tag = htonl(peer_tie_tag);

        for (count = 0; count <  num_local_Addresses; count++) {
            switch(sockunion_family(&(local_Addresses[count]))) {
                case AF_INET :
                    no_local_ipv4_addresses++;
                    break;
#ifdef HAVE_IPV6
                case AF_INET6 :
                    no_local_ipv6_addresses++;
                    break;
#endif
                default :
                    error_log(ERROR_MAJOR, "ch_enterCookieVLP: Address Type Error !");
                    break;
            }
        }
        for (count = 0; count <  num_peer_Addresses; count++) {
            switch(sockunion_family(&(peer_Addresses[count]))) {
                case AF_INET :
                    no_remote_ipv4_addresses++;
                    break;
#ifdef HAVE_IPV6
                case AF_INET6 :
                    no_remote_ipv6_addresses++;
                    break;
#endif
                default :
                    error_log(ERROR_MAJOR, "ch_enterCookieVLP: Address Type Error !");
                    break;
            }

        }
        cookie->no_local_ipv4_addresses = htons( no_local_ipv4_addresses);
        cookie->no_remote_ipv4_addresses = htons(no_remote_ipv4_addresses);

        /* TODO : IPv6 Fixes */

        cookie->no_local_ipv6_addresses = htons(no_local_ipv6_addresses);
        cookie->no_remote_ipv6_addresses = htons(no_remote_ipv6_addresses);

        wCurs = writeCursor[chunkID];

        writeCursor[chunkID] += sizeof(SCTP_cookie_param);

        event_logii(VERBOSE,
                    "Building Cookie with %u local, %u peer addresses",
                    num_local_Addresses, num_peer_Addresses);

        ch_enterIPaddresses(chunkID, local_Addresses, num_local_Addresses);
        ch_enterIPaddresses(chunkID, peer_Addresses, num_peer_Addresses);

        cookie->vlparam_header.param_length = htons(writeCursor[chunkID] - wCurs);

        cookie->sendingTime = sctp_getTime();
        cookie->cookieLifetime = cookieLifetime;

        cookie->hmac[0] = 0;
        cookie->hmac[1] = 0;
        cookie->hmac[2] = 0;
        cookie->hmac[3] = 0;

        result = signCookie((unsigned char *) &(cookie->z_side_initAck),
                            ntohs(cookie->vlparam_header.param_length) - 4, cookie->hmac);
    } else {
        error_log(ERROR_MAJOR, "ch_enterCookieVLP: chunk type not initAck");
    }

    return;
}


/* ch_initiateTag reads the initiate tag from an init or initAck */
unsigned int ch_initiateTag(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return ntohl(((SCTP_init *) chunks[chunkID])->init_fixed.init_tag);
    } else {
        error_log(ERROR_MAJOR, "ch_initiateTag: chunk type not init or initAck");
        return 0;
    }
}



/* ch_receiverWindow reads the remote receiver window from an init or initAck */
unsigned int ch_receiverWindow(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return ntohl(((SCTP_init *) chunks[chunkID])->init_fixed.rwnd);
    } else {
        error_log(ERROR_MAJOR, "ch_receiverWindow: chunk type not init or initAck");
        return 0;
    }
}



/* ch_initialTSN reads the initial TSN from an init or initAck */
unsigned int ch_initialTSN(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return ntohl(((SCTP_init *) chunks[chunkID])->init_fixed.initial_tsn);
    } else {
        error_log(ERROR_MAJOR, "ch_initialTSN: chunk type not init or initAck");
        return 0;
    }
}



/* ch_noOutStreams reads the number of output streams from an init or initAck */
unsigned short ch_noOutStreams(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return ntohs(((SCTP_init *) chunks[chunkID])->init_fixed.outbound_streams);
    } else {
        error_log(ERROR_MAJOR, "ch_noOutStreams: chunk type not init or initAck");
        return 0;
    }
}



/* ch_noInStreams reads the number of input streams from an init or initAck */
unsigned short ch_noInStreams(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return ntohs(((SCTP_init *) chunks[chunkID])->init_fixed.inbound_streams);
    } else {
        error_log(ERROR_MAJOR, "ch_noInStreams: chunk type not init or initAck");
        return 0;
    }
}


/**
 *  ch_cookieLifeTime returns the suggested cookie lifespan increment if a cookie
 *  preservative is present in a init chunk.
 */
unsigned int ch_cookieLifeTime(ChunkID chunkID)
{
    gint32 vl_param_curs;
    guint16 vl_param_total_length;
    SCTP_cookie_preservative *preserv;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        vl_param_total_length =
            ((SCTP_init *) chunks[chunkID])->chunk_header.chunk_length -
            sizeof(SCTP_chunk_header) - sizeof(SCTP_init_fixed);

        vl_param_curs = retrieveVLParamFromString(VLPARAM_COOKIE_PRESERV, &((SCTP_init *)
                                                                            chunks
                                                                            [chunkID])->
                                                  variableParams[0], vl_param_total_length);
        if (vl_param_curs >= 0) {
            /* found cookie preservative */
            preserv = (SCTP_cookie_preservative *) & ((SCTP_init *)
                                                      chunks[chunkID])->variableParams[vl_param_curs];
            return (ntohl(preserv->cookieLifetimeInc) + sci_getCookieLifeTime());
        } else {
            return sci_getCookieLifeTime();
        }
    } else {
        error_log(ERROR_MAJOR, "ch_cookieLifeTime: chunk type not init");
        return 0;
    }
}



/* ch_IPaddresses reads the IP-addresses from an init or initAck */
int ch_IPaddresses(ChunkID chunkID, union sockunion addresses[])
{
    int noOfAddresses;
    short vl_param_total_length;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        vl_param_total_length =
            ((SCTP_init *) chunks[chunkID])->chunk_header.chunk_length -
            sizeof(SCTP_chunk_header) - sizeof(SCTP_init_fixed);

        /* retrieve addresses from initAck */
        noOfAddresses = retrieveIPAddressesFromString(&((SCTP_init *)
                                                        chunks[chunkID])->variableParams[0],
                                                      vl_param_total_length, addresses);
        event_logii(VERBOSE, "Found %d addresses in %s chunk !", noOfAddresses,
                    ((chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) ? "INIT" : "INIT ACK"));

        return noOfAddresses;
    } else {
        error_log(ERROR_MINOR, "ch_IPaddresses: chunk type not init or initAck");
        return 0;
    }
}



/*
 * ch_cookieParam reads the cookie variable length parameter from an initAck
 */
SCTP_cookie_param *ch_cookieParam(ChunkID chunkID)
{
    short vl_param_curs;
    short vl_param_total_length;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return NULL;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK) {
        vl_param_total_length =
            ((SCTP_init *) chunks[chunkID])->chunk_header.chunk_length -
            sizeof(SCTP_chunk_header) - sizeof(SCTP_init_fixed);

        vl_param_curs = retrieveVLParamFromString(VLPARAM_COOKIE, &((SCTP_init *)
                                                                    chunks
                                                                    [chunkID])->variableParams
                                                  [0], vl_param_total_length);
        if (vl_param_curs >= 0) {
            /* found cookie */
            return (SCTP_cookie_param *) & ((SCTP_init *)
                                            chunks[chunkID])->variableParams[vl_param_curs];
        } else {
            /* ignore initAck message, init timer will abort */
            error_log(ERROR_MAJOR, "initAck without cookie received, message discarded");
            return NULL;
        }
    } else {
        error_log(ERROR_MINOR, "ch_cookieParam: chunk type not init or initAck");
        return NULL;
    }
}



/* ch_initFixed reads the fixed part of an init or initAck as complete structure */
SCTP_init_fixed *ch_initFixed(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return NULL;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT_ACK ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_INIT) {
        return &((SCTP_init *) chunks[chunkID])->init_fixed;
    } else {
        error_log(ERROR_MAJOR, "ch_noInStreams: chunk type not init or initAck");
        return NULL;
    }
}



/****** create and read from cookie chunk *********************************************************/

/* ch_makeCookie creates a cookie chunk.
*/
ChunkID ch_makeCookie(SCTP_cookie_param * cookieParam)
{
    SCTP_cookie_echo *cookieChunk;

    /* creat cookie chunk */
    cookieChunk = (SCTP_cookie_echo *) malloc(sizeof(SCTP_cookie_echo));

    if (cookieChunk == NULL) {
        error_log(ERROR_MAJOR, "Malloc Failed in ch_makeCookie, returning -1 !");
        return -1;
    }

    cookieChunk->chunk_header.chunk_id = CHUNK_COOKIE_ECHO;
    cookieChunk->chunk_header.chunk_flags = 0x00;
    cookieChunk->chunk_header.chunk_length = ntohs(cookieParam->vlparam_header.param_length);

    /*  copy cookie parameter EXcluding param-header into chunk            */
    /*  z_side_initAck is the first struct/data part in our cookie         */
    memcpy(&(cookieChunk->cookie), &cookieParam->z_side_initAck,
           ntohs(cookieParam->vlparam_header.param_length) - sizeof(SCTP_vlparam_header));

    enterChunk((SCTP_simple_chunk *) cookieChunk, "created cookieChunk %u ");

    return freeChunkID;
}


/*
 * ch_cookieInitFixed creates an init chunk from the fixed part of an init contained in a cookie
 *  and returns its chunkID
 */
ChunkID ch_cookieInitFixed(ChunkID chunkID)
{
    SCTP_init *initChunk;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    /* creat init chunk from init data< in cookie */
    initChunk = (SCTP_init *) malloc(sizeof(SCTP_init));
    if (initChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    /* enter fixed part of init */
    initChunk->chunk_header.chunk_id = CHUNK_INIT;
    initChunk->chunk_header.chunk_flags = 0x00;
    initChunk->chunk_header.chunk_length = sizeof(SCTP_chunk_header) + sizeof(SCTP_init_fixed);
    initChunk->init_fixed = ((SCTP_cookie_echo *) chunks[chunkID])->cookie.a_side_init;

    enterChunk((SCTP_simple_chunk *) initChunk, "created initChunk from cookie %u ");

    return freeChunkID;
}



/* ch_cookieInitAckFixed creates an initAck chunk from the fixed part of an initAck contained in a 
   cookie and returns its chunkID */
ChunkID ch_cookieInitAckFixed(ChunkID chunkID)
{
    SCTP_init *initAckChunk;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    /* creat initAck chunk from init data in cookie */
    initAckChunk = (SCTP_init *) malloc(sizeof(SCTP_init));
    if (initAckChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    /* enter fixed part of init */
    initAckChunk->chunk_header.chunk_id = CHUNK_INIT_ACK;
    initAckChunk->chunk_header.chunk_flags = 0x00;
    initAckChunk->chunk_header.chunk_length = sizeof(SCTP_chunk_header) + sizeof(SCTP_init_fixed);
    initAckChunk->init_fixed = ((SCTP_cookie_echo *) chunks[chunkID])->cookie.z_side_initAck;

    enterChunk((SCTP_simple_chunk *) initAckChunk, "created initAckChunk %u  from cookie");

    return freeChunkID;
}



/* ch_cookieIPaddresses reads the IP-addresses from a cookie */
int ch_cookieIPDestAddresses(ChunkID chunkID, union sockunion addresses[])
{
    int nAddresses;
    int vl_param_total_length;
    guint16 no_loc_ipv4_addresses, no_remote_ipv4_addresses;
    guint16 no_loc_ipv6_addresses, no_remote_ipv6_addresses;

    union sockunion temp_addresses[MAX_NUM_ADDRESSES];

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_COOKIE_ECHO) {
        no_loc_ipv4_addresses =
            ntohs(((SCTP_cookie_echo *) chunks[chunkID])->cookie.no_local_ipv4_addresses);
        no_remote_ipv4_addresses =
            ntohs(((SCTP_cookie_echo *) chunks[chunkID])->cookie.no_remote_ipv4_addresses);
        no_loc_ipv6_addresses =
            ntohs(((SCTP_cookie_echo *) chunks[chunkID])->cookie.no_local_ipv6_addresses);
        no_remote_ipv6_addresses =
            ntohs(((SCTP_cookie_echo *) chunks[chunkID])->cookie.no_remote_ipv6_addresses);

        vl_param_total_length =
            ((SCTP_cookie_echo *) chunks[chunkID])->chunk_header.chunk_length -
            COOKIE_FIXED_LENGTH - sizeof(SCTP_chunk_header);

        event_logi(VVERBOSE, " Computed total length of vparams : %d", vl_param_total_length);
        event_logii(VVERBOSE, " Num of local/remote IPv4 addresses %u / %u",
                    no_loc_ipv4_addresses, no_remote_ipv4_addresses);
        event_logii(VVERBOSE, " Num of local/remote IPv6 addresses %u / %u",
                    no_loc_ipv6_addresses, no_remote_ipv6_addresses);

        /* retrieve destination addresses from cookie */
        nAddresses = retrieveIPAddressesFromString(&((SCTP_cookie_echo *)
                                                     chunks[chunkID])->vlparams[0],
                                                   vl_param_total_length, temp_addresses);
        if (nAddresses !=
            no_loc_ipv4_addresses + no_remote_ipv4_addresses +
            no_loc_ipv6_addresses + no_remote_ipv6_addresses) {
            error_log(ERROR_MAJOR, "Parameter Problem with IP addresses in Cookie !");
        }

        memcpy(addresses, &temp_addresses[no_loc_ipv4_addresses],
               no_remote_ipv4_addresses * sizeof(union sockunion));

        if (no_remote_ipv6_addresses != 0)
            memcpy(&addresses[no_remote_ipv4_addresses],
                   &temp_addresses[no_loc_ipv4_addresses +
                                   no_remote_ipv4_addresses +
                                   no_loc_ipv6_addresses],
                   no_remote_ipv6_addresses * sizeof(union sockunion));


        return (no_remote_ipv4_addresses+no_remote_ipv6_addresses);
    } else {
        error_log(ERROR_MAJOR, "ch_cookieIPaddresses: chunk type not cookie");
        return 0;
    }
}



/* ch_staleCookie checks if this is a stale cookie and returns 0 if not and lifetime
   in msecs if it is. */
unsigned int ch_staleCookie(ChunkID chunkID)
{
    SCTP_cookie_echo *cookie_echo_chunk;
    SCTP_our_cookie *cookie_param;
    unsigned int lifetime;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return FALSE;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_COOKIE_ECHO) {
        cookie_echo_chunk = (SCTP_cookie_echo *) chunks[chunkID];
        cookie_param = &(cookie_echo_chunk->cookie);
        lifetime = sctp_getTime() - cookie_param->sendingTime;
        event_logi(INTERNAL_EVENT_0, "ch_staleCookie: lifetime = %u msecs", lifetime);

        if (lifetime > cookie_param->cookieLifetime) {
            return lifetime;
        } else
            return 0;
    } else {
        error_log(ERROR_MAJOR, "ch_staleCookie: chunk type not cookie");
        return FALSE;
    }
}

/**
 * function reads local tie tag from a received cookie echo chunk
 */
guint32 ch_CookieLocalTieTag(ChunkID chunkID)
{

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_COOKIE_ECHO) {
        return (((SCTP_cookie_echo *) chunks[chunkID])->cookie.local_tie_tag);
    } else {
        error_log(ERROR_MAJOR, "ch_CookieLocalTieTag : Not a CookieEcho chunk !");
        return 0;
    }
}

/**
 * function reads peer tie tag from a received cookie echo chunk
 */
guint32 ch_CookiePeerTieTag(ChunkID chunkID)
{

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_COOKIE_ECHO) {
        return (((SCTP_cookie_echo *) chunks[chunkID])->cookie.peer_tie_tag);
    } else {
        error_log(ERROR_MAJOR, "ch_CookiePeerTieTag : Not a CookieEcho chunk !");
        return 0;
    }
}


/**
 *    check if this is a good cookie, i.e. verify HMAC signature
 *      @return TRUE when signature is correct, else false (-1,1)
 */
boolean ch_goodCookie(ChunkID chunkID)
{
    SCTP_cookie_echo *cookie_chunk;
    SCTP_our_cookie *cookie;
    guchar cookieSignature[HMAC_LEN];
    guchar ourSignature[HMAC_LEN];
    guint16 chunklen;
    guint32 i;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return FALSE;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_COOKIE_ECHO) {
        /* this is a bit messy -- should do some cleanups here */
        cookie_chunk = (SCTP_cookie_echo *) chunks[chunkID];
        cookie = &(cookie_chunk->cookie);
        /* store HMAC */
        memcpy(cookieSignature, cookie->hmac, HMAC_LEN);

        chunklen = cookie_chunk->chunk_header.chunk_length - sizeof(SCTP_chunk_header);
        event_logi(VVERBOSE, "Got Cookie with %u bytes (incl. vlparam_header)! ", chunklen);

        signCookie((unsigned char *) cookie, chunklen, ourSignature);

        event_log(VVERBOSE, "Transmitted MD5 signature (in order to verify) : ");
        for (i = 0; i < 4; i++) {
            event_logiiii(VERBOSE, "%2.2x %2.2x %2.2x %2.2x",
                          cookieSignature[i * 4], cookieSignature[i * 4 + 1],
                          cookieSignature[i * 4 + 2], cookieSignature[i * 4 + 3]);
        }

        return (memcmp(cookieSignature, ourSignature, HMAC_LEN));

    } else {
        error_log(ERROR_MAJOR, "ch_goodCookie: chunk type not cookie");
        return FALSE;
    }
}



/****** create and read from heartbeat chunk ******************************************************/

/* ch_makeHeartbeat creates a heartbeatchunk.
*/
ChunkID ch_makeHeartbeat(unsigned int sendingTime, unsigned int pathID)
{
    SCTP_heartbeat *heartbeatChunk;

    /* creat Heartbeat chunk */
    heartbeatChunk = (SCTP_heartbeat *) malloc(sizeof(SCTP_simple_chunk));
    if (heartbeatChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    heartbeatChunk->chunk_header.chunk_id = CHUNK_HBREQ;
    heartbeatChunk->chunk_header.chunk_flags = 0;
    heartbeatChunk->chunk_header.chunk_length = sizeof(SCTP_heartbeat);
    heartbeatChunk->HB_Info.param_type = htons(VLPARAM_HB_INFO);
    heartbeatChunk->HB_Info.param_length = htons(sizeof(SCTP_heartbeat) - 4);
    heartbeatChunk->pathID = htonl((unsigned int) pathID);
    heartbeatChunk->sendingTime = htonl(sendingTime);

    enterChunk((SCTP_simple_chunk *) heartbeatChunk, "created heartbeatChunk %u ");

    return freeChunkID;
}


/* ch_HBsendingTime reads the sending time of a heartbeat.
*/
unsigned int ch_HBsendingTime(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_HBREQ ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_HBACK) {
        return ntohl(((SCTP_heartbeat *) chunks[chunkID])->sendingTime);
    } else {
        error_log(ERROR_MINOR, "ch_HBsendingTime: chunk type not heartbeat or heartbeatAck");
        return 0;
    }
}



/* ch_HBpathID reads the path heartbeat on which the heartbeat was sent.
*/
unsigned int ch_HBpathID(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_HBREQ ||
        chunks[chunkID]->chunk_header.chunk_id == CHUNK_HBACK) {
        return ntohl(((SCTP_heartbeat *) chunks[chunkID])->pathID);
    } else {
        error_log(ERROR_MINOR, "ch_HBsendingTime: chunk type not heartbeat or heartbeatAck");
        return 0;
    }
}



/***** create simple chunk **********************************************************************/

/* ch_makeSimpleChunk creates a simple chunk. It can be used for parameterless chunks like
   abort, cookieAck and shutdownAck. It can also be used for chunks that have only variable
   length parameters like the error chunks
*/
ChunkID ch_makeSimpleChunk(unsigned char chunkType, unsigned char flag)
{
    SCTP_simple_chunk *simpleChunk;

    /* creat simple chunk (used for abort, shutdownAck and cookieAck) */
    simpleChunk = (SCTP_simple_chunk *) malloc(sizeof(SCTP_simple_chunk));
    if (simpleChunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    simpleChunk->chunk_header.chunk_id = chunkType;
    simpleChunk->chunk_header.chunk_flags = flag;
    simpleChunk->chunk_header.chunk_length = 0x0004;

    enterChunk(simpleChunk, "created simpleChunk %u ");

    return freeChunkID;
}



/***** write to and read from error chunk *******************************************************/

/* enters the staleness of a cookie into an error chunk. */
void ch_enterStaleCookieError(ChunkID chunkID, unsigned int staleness)
{
    SCTP_staleCookieError *staleCE;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_ERROR) {
        if (chunkCompleted[chunkID]) {
            error_log(ERROR_MAJOR, "ch_enterStaleCookieError: chunk already completed");
            return;
        }

        staleCE =
            (SCTP_staleCookieError *) & chunks[chunkID]->simple_chunk_data[writeCursor[chunkID]];

        staleCE->vlparam_header.param_type = htons(ECC_STALE_COOKIE_ERROR);
        staleCE->vlparam_header.param_length =
            htons((unsigned short) sizeof(SCTP_staleCookieError));
        staleCE->staleness = htonl(staleness);

        writeCursor[chunkID] += (unsigned short) sizeof(SCTP_staleCookieError);
    } else {
        error_log(ERROR_MAJOR, "ch_enterStaleCookieError: chunk type not error");
    }

    return;
}



/* reads the staleness of a cookie from an error chunk. */
unsigned int ch_stalenessOfCookieError(ChunkID chunkID)
{
    short vl_param_curs;
    short vl_param_total_length;
    SCTP_staleCookieError *staleCE;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_ERROR) {
        vl_param_total_length =
            ((SCTP_simple_chunk *) chunks[chunkID])->chunk_header.chunk_length -
            sizeof(SCTP_chunk_header);

        vl_param_curs = retrieveVLParamFromString(ECC_STALE_COOKIE_ERROR, &((SCTP_simple_chunk *)
                                                                            chunks
                                                                            [chunkID])->
                                                  simple_chunk_data[0], vl_param_total_length);

        if (vl_param_curs >= 0) {
            /* found cookie staleness of cookie */
            staleCE = (SCTP_staleCookieError *) & chunks[chunkID]->simple_chunk_data[vl_param_curs];
            return ntohl(staleCE->staleness);
        } else {
            /* return 0, no effect on cookie lifetime */
            error_log(ERROR_MAJOR,
                      "ch_stalenessOfCookieError: error chunk does not contain a cookie stalenes");
            return 0;
        }
    } else {
        error_log(ERROR_MAJOR, "ch_stalenessOfCookieError: chunk type not error");
        return 0;
    }
}



void enterMissingParameterCause(unsigned short *paramTypes, short noOfParams)
{
    ;
}



void enterUnresolvableAddressCause(SCTP_ip_address address)
{
    ;
}



/***** create and read from shutdown chunk ******************************************************/


/* Creates a shutdown chunk.
*/
ChunkID ch_makeShutdown(unsigned int _cummTSNacked)
{
    SCTP_simple_chunk *shutdown_chunk;
    unsigned int *cummTSNacked;

    /* creat Shutdown chunk */
    shutdown_chunk = (SCTP_simple_chunk *) malloc(sizeof(SCTP_simple_chunk));
    if (shutdown_chunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    shutdown_chunk->chunk_header.chunk_id = CHUNK_SHUTDOWN;
    shutdown_chunk->chunk_header.chunk_flags = 0x00;
    shutdown_chunk->chunk_header.chunk_length = 0x0008;
    cummTSNacked = (unsigned int *) (&(shutdown_chunk->simple_chunk_data[0]));
    *cummTSNacked = htonl(_cummTSNacked);

    enterChunk(shutdown_chunk, "created shutdown_chunk %u ");

    return freeChunkID;
}


/* Creates a simple error chunk, with error_cause
*/
ChunkID ch_makeError(guint16 error_cause)
{
    SCTP_error_chunk *error_chunk;
    SCTP_vlparam_header *error_param;

    /* creat Shutdown chunk */
    error_chunk = malloc(sizeof(SCTP_simple_chunk));
    if (error_chunk == NULL)
        error_log_sys(ERROR_FATAL, errno);

    error_chunk->chunk_header.chunk_id = CHUNK_ERROR;
    error_chunk->chunk_header.chunk_flags = 0x00;
    error_chunk->chunk_header.chunk_length = 0x0008;

    error_param = (SCTP_vlparam_header *) & (error_chunk->data);
    error_param->param_type = htons(error_cause);
    error_param->param_length = htons(4);

    enterChunk((SCTP_simple_chunk *) error_chunk, "created shutdown_chunk %u ");
    return freeChunkID;
}


/* reads the cummulative TSN acked from a shutdown chunk.
*/
unsigned int ch_cummulativeTSNacked(ChunkID chunkID)
{
    unsigned int *cummTSNacked;

    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    if (chunks[chunkID]->chunk_header.chunk_id == CHUNK_SHUTDOWN) {
        cummTSNacked =
            (unsigned int *) (&((SCTP_simple_chunk *) chunks[chunkID])->simple_chunk_data[0]);
        return ntohl(*cummTSNacked);
    } else {
        error_log(ERROR_MAJOR, "ch_cummulativeTSNacked: chunk type not init or initAck");
        return 0;
    }
}



/****** read from, make and delete generic chunk **************************************************/

/* reads the chunks type of a chunk.
*/
unsigned char ch_chunkType(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 255;
    }

    return chunks[chunkID]->chunk_header.chunk_id;
}



/* reads the chunks length of a chunks.
*/
unsigned short ch_chunkLength(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return 0;
    }

    return chunks[chunkID]->chunk_header.chunk_length;
}



/* returns a pointer to the beginning of a simple chunk.
*/
SCTP_simple_chunk *ch_chunkString(ChunkID chunkID)
{
    if (chunks[chunkID] == NULL) {
        error_log(ERROR_MAJOR, "Invalid chunk ID");
        return NULL;
    }

    chunks[chunkID]->chunk_header.chunk_length =
        htons(chunks[chunkID]->chunk_header.chunk_length + writeCursor[chunkID]);
    chunkCompleted[chunkID] = TRUE;

    return chunks[chunkID];
}



/*
 * swaps length INSIDE the packet !!!!!!!!!!! Phew ! and puts chunk pointer
 * into the current array of chunks -- does not need ch_deleteChunk !!
 */
ChunkID ch_makeChunk(SCTP_simple_chunk * chunk)
{

    /*
     * swaps length INSIDE the packet !!!!!!!!!!! Phew ! and enters chunk
     * into the current list
     */
    chunk->chunk_header.chunk_length = ntohs(chunk->chunk_header.chunk_length);

    enterChunk(chunk, "created chunk from string %u ");

    return freeChunkID;
}


/* ch_deleteChunk removes the chunk from the array of chunks and frees the
   memory allocated for that chunk.
*/
void ch_deleteChunk(ChunkID chunkID)
{
    unsigned int cid;

    cid = chunkID;

    if (chunks[chunkID] != NULL) {
        event_logi(INTERNAL_EVENT_0, "freed chunk %u", cid);
        free(chunks[chunkID]);
        chunks[chunkID] = NULL;
    } else {
        error_log(ERROR_MAJOR, "chunk already freed");
    }
}



/* ch_forgetChunk removes the chunk from the array of chunks without freeing the
   memory allocated for that chunk.
   This is used in the following cases:
   - the caller wants to keep the chunk for retransmissions.
   - the chunk was created with ch_makeChunk and the pointer to the chunk points
     into an SCTP-message, which was allocated as a whole. In this case the chunk
     can not be freed here.
*/
void ch_forgetChunk(ChunkID chunkID)
{
    unsigned int cid;

    cid = chunkID;

    if (chunks[chunkID] != NULL) {
        chunks[chunkID] = NULL;
        event_logi(INTERNAL_EVENT_0, "forgot chunk %u", cid);
    } else {
        error_log(ERROR_MAJOR, "chunk already forgotten");
    }
}
